(* The function phi(z, qbar) from the paper *)
Phi[z_,qbar_] := 1 + (1/3+z+qbar*z)/(1/3 - 2*z - qbar*z - (1/3+z)*(2*qbar*z)/(1/3-2*z))
	
(* Solves the LP from the paper for the given set of parameters *)
LP[z_, t_, version_] := Maximize[{

	(* ========================  Objective  =============================*)
	(1 - t) * (2/3 + 2*z - q*z) 
	+ t*(
		Phi[z, 3]
		*
		(sb + (1/3 + z) * ss + (2/3 + 2*z)*so)
		+
		 ob + (1/3 + z) * os + (2/3 + 2*z)*oo
	),

	(* ========================  Constraints  ===========================*)

	(* Constraints from the paper *)
	bb + bs + bo >= (1/3 + z - qbar*z - (1/3 + z)*(ob + os + oo))/(2/3 - z),
	bb + sb + ob >= (1/3 + z - q*z - (1/3 + z)*(bo + so + oo))/(2/3 - z),
	sb + ss + so >= (1/3 - 2*z - qbar*z - (1/3 + z)*(ob + os + oo))/(2/3 - z),
	bs + ss + os >= (1/3 - 2*z - q*z - (1/3 + z)*(bo + so + oo))/(2/3 - z),
	ob + os + oo <= 2*qbar*z/(1/3 - 2*z),
	bo + so + oo <= 2*q*z/(1/3 - 2*z),

	(* Either bs = 0 or sb = 0 *)
	version == 0,
	
	(* Sum to one *)
	bb + bs + bo + sb + ss + so + ob + os + oo == 1,

	(* qbar constraints *)
	qbar <= 3,
	qbar <= q,

	(* Non-negativities *)
	bb >= 0,
	bs >= 0,
	bo >= 0,
	sb >= 0,
	ss >= 0,
	so >= 0,
	ob >= 0,
	os >= 0,
	oo >= 0,
	q >= 0,
	qbar >= 0

	},

	(* ========================  Variables  ============================ *)
	{bb, bs, bo, sb, ss, so, ob, os, oo, q, qbar}
]

(* For a given z and k, this function uses grid search to find the t that
 minimizes the objective function of LP(z, t, k) *)
FindT[z_, version_] := Module[{t, tstep, best, tbest, outcome},
	t = 0;
	tstep = 1/1000;
	best = 1;
	tbest = 0;
	While[t <= 2/10,
		outcome = LP[z, t, version][[1]];
		If[outcome < best, best = outcome; tbest = t];
		t += tstep
	];
	{best, tbest}
]

(* This function uses grid search to find the best witness (z, t_0, t_1). We
   have chosen starting parameters that give the result from the paper in a
   reasonable amount of time *)
Module[{z, zstep, zbest, tbest, outcome},
	(* Starting z. This can be reduced to 0 if desired, but the program will
	 take a very long time if zstep is not increased. *)
	z = 59137/10000000;
	zstep = 1/1000000000;
	zbest = 0;
	t1best = -1;
	t2best = -1;
	(* The upper bound on z can be increased to any number strictly less than
	 (13-3*Sqrt[17])/24, but the program will take a very long time if zstep is
	 not increased *)
	While[z <= 59138/10000000,
		z = z + zstep;
		out1 = FindT[z, sb];
		out2 = FindT[z, bs];
		outcome = Max[out1[[1]], out2[[1]]];
		If[outcome <= 2/3-z, zbest = z; t1best = out1[[2]]; t2best = out2[[2]]]
	];
	(* Print out the exact rational z *)
	Print[zbest]
	(* Print out the witness in decimal form *)
	Print[N[zbest]];
	Print[N[t2best]];
	Print[N[t1best]];
]
